! Fortran 95/2003 Exercise 3
program GameOfLife
  ! 2D cellular automaton ("Game of Life")
  ! (c) Pekka Manninen, Jussi Enkovaara, CSC - IT Center for Science Ltd
  use gol_io, only : read_data, draw
  implicit none
  integer, parameter :: &
       freq = 10           ! after how many iterations we print the board
  integer, dimension(:,:), pointer :: &
       board_1, board_2     ! GoL boards
  integer :: &
       height, width, &     ! board dimensions
       maxiter, i, &
       t0, t1, cnt, cnt_max

  ! read in the user input (iterations, board dimensions)
  call read_data(maxiter, height, width)
  ! initialize boards
  call init_board(board_1, board_2, height, width)
  ! develop the automaton
  call system_clock(t0, cnt, cnt_max)
  do i = 1, maxiter
     call update(board_1, board_2, height, width)
     if (mod(i,freq)==0) call draw(board_1, height, width, i)
  end do
  call system_clock(t1,cnt,cnt_max)
  write(*,'(A,I4,A,F6.2,A)') 'Completed ', maxiter, ' iterations in ', &
       real(t1-t0) / real(cnt), ' seconds.'

contains

  subroutine init_board(board_1, board_2, height, width)
    implicit none
    integer, dimension(:,:), pointer :: board_1, board_2
    integer, intent(in) :: height, width

    allocate(board_1(height, width))
    allocate(board_2(height, width))
    ! start from a plus-sign shaped pattern
    board_1(:,:) = 0
    board_2(:,:) = 0
    board_1(height/2,:) = 1
    board_1(:,width/2) = 1
    call draw(board_1, height, width, 0)
  end subroutine init_board

  subroutine update(board_1, board_2, heigth, width)
    implicit none
    integer, parameter :: ALIVE = 1, DEAD = 0
    integer, dimension(:,:), pointer :: board_1, board_2
    integer, dimension(:,:), pointer :: temp_board
    integer, intent(in) :: heigth, width
    integer :: i, j, nsum, ip, im, jp, jm

    do j = 1, width
       ! periodic boundaries
       jm = j - 1
       jp = j + 1
       if (j == 1)     jm = width
       if (j == width) jp = 1
       do i = 1, height
          im = i - 1
          ip = i + 1
          if (i == 1)      im = height
          if (i == height) ip = 1
          ! determine the sum of neighboring elements
          nsum = board_1(im, jm) + board_1(im, j) + board_1(im, jp) + &
               & board_1(i,  jm) +                  board_1(i,  jp) + &
               & board_1(ip, jm) + board_1(ip, j) + board_1(ip, jp)
          if (nsum == 2) then
             board_2(i,j) = board_1(i,j)
          else if (nsum == 3) then
             board_2(i,j) = ALIVE
          else
             board_2(i,j) = DEAD
          end if
       end do
    end do

    ! swap board contents (such that the old <-> new)
    temp_board => board_1
    board_1 => board_2
    board_2 => temp_board
  end subroutine update

end program GameOfLife
