-! CSC Summer School 2012
! Introduction to Fortran 95/2003 Exercise 3 c
program GameOfLife
! 2D cellular automaton ("Game of Life")
! (c) Pekka Manninen, Jussi Enkovaara, CSC - IT Center for Science Ltd
  use gol_io, only : GoL_board, read_data_dd, draw_dd
  implicit none
  integer, parameter :: &
      freq = 10           ! after how many iterations we print the board
  type(GoL_board) :: board_1, board_2     ! GoL boards
  integer :: &
      maxiter, i, &
      t0, t1, cnt, cnt_max

  ! read in the user input (iterations, board dimensions)
  call read_data_dd(board_1, maxiter)
  ! initialize boards
  call init_board(board_1, board_2)
  ! develop the automaton
  call system_clock(t0,cnt,cnt_max)
  do i = 1, maxiter
    call update(board_1, board_2, i)
    if (mod(i,freq) == 0) call draw_dd(board_1)
  end do
  call system_clock(t1, cnt, cnt_max)
  write(*,'(A,I4,A,F6.2,A)') 'Completed ', maxiter, ' iterations in ', &
       real(t1-t0) / real(cnt), ' seconds.'

contains

  subroutine init_board(board_1, board_2)
    implicit none
    type(GoL_board), intent(inout) :: board_1, board_2

    ! copy the dimensions, allocate boards
    board_2%height = board_1%height
    board_2%width = board_1%width
    allocate(board_1%board(board_1%height, board_1%width))
    allocate(board_2%board(board_2%height, board_2%width))

    ! start from a plus-sign shaped pattern
    board_1%board(:,:) = 0
    board_2%board(:,:) = 0
    board_1%board(board_1%height/2,:) = 1
    board_1%board(:,board_2%width/2) = 1
    board_1%iter = 0
    board_2%iter = 0
    call draw_dd(board_1)
  end subroutine init_board

  subroutine update(board_1, board_2, it)
    implicit none
    integer, parameter :: ALIVE = 1, DEAD = 0
    type(GoL_board) :: board_1, board_2
    integer, intent(in) :: it
    integer, dimension(:,:), pointer :: temp_board
    integer :: i, j, nsum, ip, im, jp, jm

    do j = 1, board_1%width
       ! periodic boundaries
       jm = j - 1
       jp = j + 1
       if (j == 1) jm = board_1%width
       if (j == board_1%width) jp = 1
       do i = 1, board_1%height
          im = i - 1
          ip = i + 1
          if (i == 1) im = board_1%height
          if (i == board_1%height) ip = 1
          ! determine the sum of neighboring elements
          nsum = board_1%board(im,jm) + board_1%board(im,j) + board_1%board(im,jp) + &
               & board_1%board(i, jm) +                       board_1%board(i, jp) + &
               & board_1%board(ip,jm) + board_1%board(ip,j) + board_1%board(ip,jp)
          if (nsum == 2) then
             board_2%board(i,j) = board_1%board(i,j)
          else if (nsum == 3) then
             board_2%board(i,j) = ALIVE
          else
             board_2%board(i,j) = DEAD
          end if
       end do
    end do

 ! swap board contents (such that the old <-> new)
    temp_board => board_1%board
    board_1%board => board_2%board
    board_2%board => temp_board

 ! update the iteration counters
    board_1%iter = it
    board_2%iter = it
  end subroutine update

end program GameOfLife
